#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Thu Jan 11 2024 14:12:29 by codeskyblue
"""


from __future__ import annotations

import logging
import os
import shlex

import click
from pymobiledevice3.cli.cli_common import print_json
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments.device_info import DeviceInfo
from pymobiledevice3.services.dvt.instruments.process_control import ProcessControl
from pymobiledevice3.services.installation_proxy import InstallationProxyService

from tidevice3.api import proclist
from tidevice3.cli.cli_common import cli, pass_rsd, pass_service_provider
from tidevice3.exceptions import FatalError
from tidevice3.utils.common import print_dict_as_table
from tidevice3.utils.download import download_file, is_hyperlink

logger = logging.getLogger(__name__)

@cli.group()
def app():
    """app related commands"""
    pass


@app.command("install")
@click.argument("path_or_url")
@pass_service_provider
def app_install(service_provider: LockdownClient, path_or_url: str):
    """install given .ipa or url"""
    if is_hyperlink(path_or_url):
        ipa_path = download_file(path_or_url)
    elif os.path.isfile(path_or_url):
        ipa_path = path_or_url
    else:
        raise ValueError("local file not found", path_or_url)
    InstallationProxyService(lockdown=service_provider).install_from_local(ipa_path)


@app.command("list")
@click.option('app_type', '-t', '--type', type=click.Choice(['System', 'User', 'Hidden', 'Any']), default='Any',
              help='include only applications of given type')
@click.option("--calculate-sizes/--no-calculate-size", default=False)
@pass_service_provider
def app_list(service_provider: LockdownClient, app_type: str, calculate_sizes: bool):
    """list installed apps"""
    app_infos = InstallationProxyService(lockdown=service_provider).get_apps(app_type, calculate_sizes=calculate_sizes)
    print_dict_as_table(app_infos.values(), ["CFBundleIdentifier", "CFBundleDisplayName", "CFBundleVersion", "CFBundleShortVersionString"])


@app.command("uninstall")
@click.argument("bundle_identifier")
@pass_service_provider
def app_uninstall(service_provider: LockdownClient, bundle_identifier: str):
    """uninstall application"""
    InstallationProxyService(lockdown=service_provider).uninstall(bundle_identifier)


@app.command("launch")
@click.argument("arguments", type=click.STRING)
@click.option("--kill-existing/--no-kill-existing", default=True, help="Whether to kill an existing instance of this process")
@click.option("--suspended", is_flag=True, help="Same as WaitForDebugger")
@click.option("--env", multiple=True, type=click.Tuple((str, str)), help="Environment variables to pass to process given as a list of key value")
@click.option("--stream", is_flag=True)
@pass_rsd
def app_launch(service_provider, arguments: str, kill_existing: bool, suspended: bool, env: tuple, stream: bool):
    """launch application"""
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        process_control = ProcessControl(dvt)
        parsed_arguments = shlex.split(arguments)
        pid = process_control.launch(
            bundle_id=parsed_arguments[0],
            arguments=parsed_arguments[1:],
            kill_existing=kill_existing,
            start_suspended=suspended,
            environment=dict(env),
        )
        print(f"Process launched with pid {pid}")
        while stream:
            for output_received in process_control:
                logging.getLogger(f"PID:{output_received.pid}").info(
                    output_received.message.strip()
                )


@app.command("kill")
@click.argument("pid", type=click.INT)
@pass_rsd
def app_kill(service_provider, pid: int):
    """kill application"""
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        process_control = ProcessControl(dvt)
        process_control.kill(pid)


@app.command("ps")
@click.option('--json/--no-json', default=False)
@pass_rsd
def app_ps(service_provider: LockdownClient, json: bool):
    """list running processes"""
    if service_provider.product_version < "17":
        logger.warning('iOS<17 have FD leak, which will cause an error when calling round more than 250 times.')
    processes = list(proclist(service_provider))
    processes = [p.model_dump(exclude_none=True) for p in processes if p.isApplication]
    if json:
        print_json(processes)
    else:
        print_dict_as_table(processes, ["pid", "name", "bundleIdentifier", "realAppName"])


@app.command("foreground")
@pass_rsd
def app_foreground(service_provider: LockdownClient):
    """show foreground running app, requires iOS>=17"""
    if service_provider.product_version < "17":
        raise FatalError("iOS<17 not supported")
    for p in proclist(service_provider):
        if p.foregroundRunning:
            print(p.bundleIdentifier, f"pid:{p.pid}")


@app.command("info")
@click.argument("bundle_identifier")
@pass_rsd
def app_info(service_provider: LockdownClient, bundle_identifier: str):
    with InstallationProxyService(lockdown=service_provider) as iproxy:
        apps = iproxy.get_apps(bundle_identifiers=[bundle_identifier])
        print_json(apps)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Jan 05 2024 23:09:36 by codeskyblue
"""

from __future__ import annotations

import re

__all__ = ["download_file", "is_hyperlink", "DownloadError"]

import base64
import hashlib
import logging
import os
import pathlib
import shutil
import time
from typing import Optional, Union

import requests
from requests.structures import CaseInsensitiveDict

from tidevice3.exceptions import DownloadError

logger = logging.getLogger(__name__)

CACHE_DOWNLOAD_SUFFIX = ".t3-download-cache"
DEFAULT_DOWNLOAD_TIMEOUT = 600  # 10 minutes

StrOrPathLike = Union[str, pathlib.Path]


def md5sum(filepath: StrOrPathLike) -> str:
    """return md5sum of given file"""
    m = hashlib.md5()
    with open(filepath, "rb") as f:
        while True:
            data = f.read(1<<20)
            if not data:
                break
            m.update(data)
    return m.hexdigest()


class RemoteFileInfo:
    content_md5: str | None = None
    content_length: int = 0
    accept_ranges: bool = False


def get_remote_file_info(headers: CaseInsensitiveDict) -> RemoteFileInfo:
    """
    Get remote file info, such as content-length, content-md5, accept-ranges
    """
    info = RemoteFileInfo()
    info.content_length = int(headers.get("content-length", 0))
    info.accept_ranges = headers.get("accept-ranges") == "bytes"
    md5_base64 = headers.get("content-md5")
    if md5_base64:
        try:
            content_md5 = base64.b64decode(md5_base64).hex()
            if len(content_md5) == 32:
                info.content_md5 = content_md5
        except:
            pass
    return info


def check_if_already_downloaded(
    filepath: pathlib.Path, remote_file_info: RemoteFileInfo
) -> bool:
    if not filepath.exists():
        return False
    if filepath.stat().st_size != remote_file_info.content_length:
        return False
    if remote_file_info.content_md5:
        if md5sum(filepath) != remote_file_info.content_md5:
            return False
    return True


def update_file_mtime(filepath: pathlib.Path):
    """update file mtime to avoid flie being deleted by clean script"""
    _atime, _mtime = (time.time(), time.time())
    os.utime(filepath, (_atime, _mtime))


def download_file_from_range(
    url: str, filepath: pathlib.Path, bytes_start: int, timeout: float
):
    r = make_request_get_stream(
        url, timeout, headers={"Range": f"bytes={bytes_start}-"}
    )
    with filepath.open("ab") as f:
        shutil.copyfileobj(r.raw, f)


def get_bytes_start(
    tmpfpath: pathlib.Path, remote_file_info: RemoteFileInfo
) -> Optional[int]:
    if (
        remote_file_info.accept_ranges
        and tmpfpath.exists()
        and tmpfpath.stat().st_mtime > time.time() - 60
    ):
        return tmpfpath.stat().st_size


def make_request_get_stream(
    url: str, timeout: float, headers: dict = None
) -> requests.Response:
    r = requests.get(url, stream=True, timeout=timeout, headers=headers)
    try:
        r.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise DownloadError(e, url)
    return r


def is_hyperlink(url: str) -> bool:
    return url.startswith("http://") or url.startswith("https://")


def guess_filename_from_url(url: str, headers: CaseInsensitiveDict = {}) -> str:
    """
    Guess filename from url and headers
    """
    filename = url.split("/")[-1]
    filename = re.sub(r"\?.*$", "", filename)
    if "content-disposition" in headers:
        for part in headers["content-disposition"].split(";"):
            if part.strip().startswith("filename="):
                filename = part.split("=")[-1].strip('"')
    return filename


def download_file(
    url: str, filepath: StrOrPathLike | None = None, timeout: float = DEFAULT_DOWNLOAD_TIMEOUT
) -> pathlib.Path:
    """
    Download file from given url to filepath

    :param url: url to download
    :param filepath: local file path
    :param timeout: timeout in seconds

    raise DownloadError if download failed
    """
    if not is_hyperlink(url):
        raise DownloadError("only support http/https url", url)
    
    logger.info("download from url: %s", url)
    r = make_request_get_stream(url, timeout)

    if filepath is None:
        filepath = guess_filename_from_url(url, r.headers)
    filepath = pathlib.Path(filepath)
    tmpfpath = pathlib.Path(
        str(filepath) + CACHE_DOWNLOAD_SUFFIX
    )  # 文件先下载到这里，等检查完后再Move过去

    remote_file_info = get_remote_file_info(r.headers)
    if check_if_already_downloaded(filepath, remote_file_info):
        logger.debug("use cached asset: %s", filepath)
        update_file_mtime(filepath)
        return filepath

    bytes_start = get_bytes_start(tmpfpath, remote_file_info)
    if bytes_start:
        logger.debug("resume download from %s", bytes_start)
        download_file_from_range(url, tmpfpath, bytes_start, timeout)
    else:
        with tmpfpath.open("wb") as f:
            shutil.copyfileobj(r.raw, f)
    if not check_if_already_downloaded(tmpfpath, remote_file_info):
        tmpfpath.unlink(missing_ok=True)
        raise DownloadError("download file not complete", url)
    os.rename(tmpfpath, filepath)
    return filepath



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Jan 09 2024 19:09:47 by codeskyblue
"""

from __future__ import annotations

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.diagnostics import DiagnosticsService

from tidevice3.cli.cli_common import cli, pass_service_provider


@cli.command()
@pass_service_provider
def reboot(service_provider: LockdownClient):
    """reboot device"""
    with DiagnosticsService(service_provider) as diagnostics:
        diagnostics.restart()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Jan 05 2024 18:10:32 by codeskyblue
"""
from __future__ import annotations

import collections
from functools import update_wrapper

import click
from pymobiledevice3.cli.cli_common import USBMUX_OPTION_HELP

from tidevice3.api import connect_service_provider


class OrderedGroup(click.Group):
    def __init__(self, name=None, commands=None, *args, **attrs):
        super(OrderedGroup, self).__init__(name, commands, *args, **attrs)
        #: the registered subcommands by their exported names.
        self.commands = commands or collections.OrderedDict()

    def list_commands(self, ctx):
        return self.commands


@click.group(cls=OrderedGroup, context_settings=dict(help_option_names=["-h", "--help"]))
@click.option("-u", "--udid", default=None, help="udid of device")
@click.option("usbmux_address", "--usbmux", help=USBMUX_OPTION_HELP)
@click.pass_context
def cli(ctx: click.Context, udid: str, usbmux_address: str):
    ctx.ensure_object(dict)
    ctx.obj['udid'] = udid
    ctx.obj['usbmux_address'] = usbmux_address


def pass_service_provider(func):
    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        udid = ctx.obj['udid']
        usbmux_address = ctx.obj['usbmux_address']
        service_provider = connect_service_provider(udid, force_usbmux=True, usbmux_address=usbmux_address)
        with service_provider:
            return ctx.invoke(func, service_provider, *args, **kwargs)
    return update_wrapper(new_func, func)


def pass_rsd(func):
    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        udid = ctx.obj['udid']
        usbmux_address = ctx.obj['usbmux_address']
        service_provider = connect_service_provider(udid=udid, usbmux_address=usbmux_address)
        with service_provider:
            return ctx.invoke(func, service_provider, *args, **kwargs)
    return update_wrapper(new_func, func)


CLI_GROUPS = ["list", "developer", "screenshot", "screenrecord", "fsync", "app", "reboot", "tunneld", "runwda", "exec"]
for group in CLI_GROUPS:
    __import__(f"tidevice3.cli.{group}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Jan 09 2024 16:56:42 by codeskyblue
"""
from __future__ import annotations

import functools
import threading
import unicodedata


def threadsafe_function(fn):
    """
    A decorator to make thread-safe functions by using a lock.

    Args:
    fn (function): The function to be decorated.

    Returns:
    function: The decorated thread-safe function.
    """
    lock = threading.Lock()

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        with lock:
            return fn(*args, **kwargs)

    return wrapper


def unicode_len(s: str) -> int:
    """ printable length of string """
    length = 0
    for char in s:
        if unicodedata.east_asian_width(char) in ('F', 'W'):
            length += 2
        else:
            length += 1
    return length


def ljust(s, length: int):
    s = str(s)
    return s + ' ' * (length - unicode_len(s))


def print_dict_as_table(dict_values: list[dict], headers: list[str], sep: str = " "):
    """
    Output as format
    ----------------------------------------
    Identifier                DeviceName ProductType ProductVersion ConnectionType
    00000000-1234567890123456 MIMM       iPhone13,3  17.2           USB
    """
    header_with_lengths = []
    for header in headers:
        if dict_values:
            max_len = max([unicode_len(str(item.get(header, ""))) for item in dict_values])
        else:
            max_len = 0
        header_with_lengths.append((header, max(max_len, unicode_len(header))))
    rows = []
    # print header
    for header, _len in header_with_lengths:
        rows.append(ljust(header, _len))
    print(sep.join(rows).rstrip())
    # print rows
    for item in dict_values:
        rows = []
        for header, _len in header_with_lengths:
            rows.append(ljust(item.get(header, ""), _len))
        print(sep.join(rows).rstrip())
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Mon Jan 29 2024 14:15:52 by codeskyblue
"""


import logging
import threading
import time
import typing

import click
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.dvt.testmanaged.xcuitest import XCUITestService
from pymobiledevice3.services.installation_proxy import InstallationProxyService
from pymobiledevice3.tcp_forwarder import UsbmuxTcpForwarder

from tidevice3.cli.cli_common import cli, pass_rsd

logger = logging.getLogger(__name__)


def guess_wda_bundle_id(service_provider: LockdownClient) -> typing.Optional[str]:
    app_infos = InstallationProxyService(lockdown=service_provider).get_apps('User')
    wda_bundle_ids = []
    for bundle_id in app_infos.keys():
        if bundle_id.endswith(".xctrunner"):
            wda_bundle_ids.append(bundle_id)
    wda_bundle_ids.sort(key=lambda x: x.find('WebDriverAgentRunner'), reverse=True)
    if not wda_bundle_ids:
        return None
    return wda_bundle_ids[0]


@cli.command("runwda")
@click.option('--bundle-id', default=None, help="WebDriverAgent bundle id")
@click.option("--src-port", default=8100, help="WebDriverAgent listen port")
@click.option("--dst-port", default=8100, help="local listen port")
@pass_rsd
def cli_runwda(service_provider: LockdownClient, bundle_id: str, src_port: int, dst_port: int):
    """run WebDriverAgent"""
    if not bundle_id:
        bundle_id = guess_wda_bundle_id(service_provider)
        if not bundle_id:
            raise ValueError("WebDriverAgent not found")
    
    def tcp_forwarder():
        logger.info("forwarder started, listen on %s", dst_port)
        forwarder = UsbmuxTcpForwarder(service_provider.udid, dst_port, src_port)
        forwarder.start()
    
    def xcuitest():
        XCUITestService(service_provider).run(bundle_id)

    thread1 = threading.Thread(target=tcp_forwarder, daemon=True)
    thread2 = threading.Thread(target=xcuitest, daemon=True)
    thread1.start()
    thread2.start()

    while thread1.is_alive() and thread2.is_alive():
        time.sleep(0.1)
    logger.info("Program exited")
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Jan 05 2024 18:10:01 by codeskyblue
"""
from __future__ import annotations

import click
from pymobiledevice3.cli.cli_common import print_json

from tidevice3.api import list_devices
from tidevice3.cli.cli_common import cli
from tidevice3.utils.common import print_dict_as_table


@cli.command(name="list")
@click.option("-u", "--usb", is_flag=True, help="show only usb devices")
@click.option("-n", "--network", is_flag=True, help="show only network devices")
@click.option("--json", is_flag=True, help="output as json format")
@click.pass_context
def cli_list(ctx: click.Context, usb: bool, network: bool, json: bool):
    """list connected devices"""
    usbmux_address = ctx.obj["usbmux_address"]
    devices = list_devices(usb, network, usbmux_address)
    if json:
        print_json([d.model_dump() for d in devices])
    else:
        headers = ["Identifier", "DeviceName", "ProductType", "ProductVersion", "ConnectionType"]
        print_dict_as_table([d.model_dump() for d in devices], headers)






from __future__ import annotations

import datetime
import io
from typing import Iterator, Optional

import requests
from PIL import Image
from pydantic import BaseModel
from pymobiledevice3.lockdown import LockdownClient, create_using_usbmux, usbmux
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments.device_info import DeviceInfo
from pymobiledevice3.services.dvt.instruments.screenshot import Screenshot
from pymobiledevice3.services.screenshot import ScreenshotService

from tidevice3.exceptions import FatalError


class DeviceShortInfo(BaseModel):
    BuildVersion: str
    ConnectionType: Optional[str]
    DeviceClass: str
    DeviceName: str
    Identifier: str
    ProductType: str
    ProductVersion: str


class ProcessInfo(BaseModel):
    isApplication: bool
    pid: int
    name: str
    realAppName: str
    startDate: datetime.datetime
    bundleIdentifier: Optional[str] = None
    foregroundRunning: Optional[bool] = None


def list_devices(
    usb: bool = True, network: bool = False, usbmux_address: Optional[str] = None
) -> list[DeviceShortInfo]:
    connected_devices = []
    for device in usbmux.list_devices(usbmux_address=usbmux_address):
        udid = device.serial

        if usb and not device.is_usb:
            continue

        if network and not device.is_network:
            continue

        lockdown = create_using_usbmux(
            udid,
            autopair=False,
            connection_type=device.connection_type,
            usbmux_address=usbmux_address,
        )
        info = DeviceShortInfo.model_validate(lockdown.short_info)
        connected_devices.append(info)
    return connected_devices


DEFAULT_TIMEOUT = 60

def connect_service_provider(udid: Optional[str], force_usbmux: bool = False, usbmux_address: Optional[str] = None) -> LockdownServiceProvider:
    """Connect to device and return LockdownServiceProvider"""
    lockdown = create_using_usbmux(serial=udid, usbmux_address=usbmux_address)
    if force_usbmux:
        return lockdown
    if lockdown.product_version >= "17":
        return connect_remote_service_discovery_service(lockdown.udid)
    return lockdown


def connect_remote_service_discovery_service(udid: str, tunneld_url: str = 'http://localhost:5555') -> RemoteServiceDiscoveryService:
    try:
        resp = requests.get(tunneld_url, timeout=DEFAULT_TIMEOUT)
        tunnels = resp.json()
        ipv6_address = tunnels.get(udid)
        if ipv6_address is None:
            raise FatalError("tunneld not ready for device", udid)
        rsd = RemoteServiceDiscoveryService(ipv6_address)
        rsd.connect()
        return rsd
    except requests.RequestException:
        raise FatalError("Please run `sudo t3 tunneld` first")
    except (TimeoutError, ConnectionError):
        raise FatalError("RemoteServiceDiscoveryService connect failed")

def iter_screenshot(service_provider: LockdownClient) -> Iterator[bytes]:
    if int(service_provider.product_version.split(".")[0]) >= 17:
        with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
            screenshot_service = Screenshot(dvt)
            while True:
                yield screenshot_service.get_screenshot()
    else:
        screenshot_service = ScreenshotService(service_provider)
        while True:
            yield screenshot_service.take_screenshot()


def screenshot_png(service_provider: LockdownClient) -> bytes:
    """ get screenshot as png data """
    it = iter_screenshot(service_provider)
    png_data = next(it)
    it.close()
    return png_data


def screenshot(service_provider: LockdownClient) -> Image.Image:
    """ get screenshot as PIL.Image.Image """
    png_data = screenshot_png(service_provider)
    return Image.open(io.BytesIO(png_data)).convert("RGB")


def proclist(service_provider: LockdownClient) -> Iterator[ProcessInfo]:
    """ list running processes"""
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        processes = DeviceInfo(dvt).proclist()
        for process in processes:
            if 'startDate' in process:
                process['startDate'] = str(process['startDate'])
                yield ProcessInfo.model_validate(process)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Jan 05 2024 18:10:01 by codeskyblue
"""
import logging

from packaging.version import Version
from pymobiledevice3.common import get_home_folder
from pymobiledevice3.exceptions import AlreadyMountedError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.amfi import AmfiService
from pymobiledevice3.services.mobile_image_mounter import auto_mount

from tidevice3.cli.cli_common import cli, pass_service_provider

logger = logging.getLogger(__name__)


@cli.command(name="developer")
@pass_service_provider
def cli_developer(service_provider: LockdownClient):
    """ enable developer mode """
    if Version(service_provider.product_version) >= Version("16"):
        if not service_provider.developer_mode_status:
            logger.info('enable developer mode')
            AmfiService(service_provider).enable_developer_mode()
        else:
            logger.info('developer mode already enabled')
    
    try:
        xcode = get_home_folder() / 'Xcode.app'
        xcode.mkdir(parents=True, exist_ok=True)
        auto_mount(service_provider, xcode=xcode)
        logger.info('mount developer image')
    except AlreadyMountedError:
        logger.info('developer image already mounted')
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Jan 05 2024 23:11:55 by codeskyblue
"""

class BaseException(Exception):
    pass

class DownloadError(BaseException):
    pass

class FatalError(BaseException):
    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Jan 09 2024 14:05:31 by codeskyblue
"""

from __future__ import annotations

import datetime
import pathlib
import posixpath
from functools import update_wrapper
from typing import List

import click
from pydantic import BaseModel
from pymobiledevice3.exceptions import AfcException, AfcFileNotFoundError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.afc import AfcService
from pymobiledevice3.services.house_arrest import HouseArrestService

from tidevice3.cli.cli_common import cli, pass_service_provider
from tidevice3.exceptions import FatalError


def pass_afc(func):
    @pass_service_provider
    @click.pass_context
    def new_func(ctx: click.Context, service_provider: LockdownClient, *args, **kwargs):
        if ctx.obj['bundle_id']:
            afc = HouseArrestService(lockdown=service_provider,
                                     bundle_id=ctx.obj['bundle_id'],
                                     documents_only=ctx.obj["documents"])
        else:
            afc = AfcService(lockdown=service_provider)
        
        with afc:
            try:
                return ctx.invoke(func, afc, *args, **kwargs)
            except AfcFileNotFoundError as e:
                raise FatalError(f"File not found: {e}")
    return update_wrapper(new_func, func)


@cli.group()
@click.option("-B", "--bundle-id", help="bundle id of app")
@click.option('--documents', is_flag=True)
@click.pass_context
def fsync(ctx: click.Context, bundle_id: str, documents: bool):
    """file sync"""
    ctx.ensure_object(dict)
    ctx.obj["bundle_id"] = bundle_id
    ctx.obj["documents"] = documents



@fsync.command(name="ls")
@click.argument("remote_path", required=True)
@click.option('-r', '--recursive', is_flag=True)
@pass_afc
def list(afc: AfcService, remote_path: str, recursive: bool):
    """ perform a dirlist rooted at /var/mobile/Media """
    items: List[FileInfo] = []
    for name in afc.listdir(remote_path):
        try:
            file_info = stat_file(afc, posixpath.join(remote_path, name))
            items.append(file_info)
        except AfcException as e:
            pass
    items.sort(key=lambda x: [x.is_dir(), x.mtime], reverse=True)
    for item in items:
        name = item.name + ("/" if item.is_dir() else "")
        size = byte2humansize(item.size)
        click.echo(f"{item.ifmt[2:]}\t{size}\t{name}")


@fsync.command(name="rm")
@click.argument("path", default="/")
@pass_afc
def remove(afc: AfcService, path: str):
    """ remove a file rooted at /var/mobile/Media """
    afc.rm(path)
    click.echo(path)


@fsync.command(name="push")
@click.argument('local_file', type=click.File('rb'))
@click.argument('remote_file', type=click.Path(exists=False))
@pass_afc
def afc_push(afc: AfcService, local_file, remote_file):
    """ push local file into /var/mobile/Media """
    finfo = stat_file(afc, remote_file)
    if finfo.is_dir():
        remote_file = posixpath.join(remote_file, local_file.name)
    afc.set_file_contents(remote_file, local_file.read())


@fsync.command('pull')
@click.option("-f", "--force", is_flag=True, help="force overwrite")
@click.argument('remote_file', type=click.Path(exists=False))
@click.argument('local_file', default="./", type=click.Path(exists=False, path_type=pathlib.Path))
@pass_afc
def afc_pull(afc: AfcService, remote_file, local_file: pathlib.Path, force: bool):
    """ pull remote file from /var/mobile/Media """
    if local_file.is_dir():
        local_file /= posixpath.basename(remote_file)
    
    if local_file.exists():
        if not force:
            raise click.BadParameter("local_file already exists")
    elif not local_file.parent.exists():
        raise click.BadParameter("local_file's parent not exists")
    
    finfo = stat_file(afc, remote_file)
    if finfo.is_dir():
        raise click.BadParameter("remote_file is a directory")
    local_file.write_bytes(afc.get_file_contents(remote_file))
    click.echo(f"remote:{remote_file} -> local:{local_file}")


class FileInfo(BaseModel):
    name: str
    size: int
    mtime: datetime.datetime
    ifmt: str

    def is_dir(self) -> bool:
        return self.ifmt == "S_IFDIR"


def stat2fileinfo(info: dict) -> FileInfo:
    # {'st_size': 326, 'st_blocks': 8, 'st_nlink': 1, 'st_ifmt': 'S_IFREG',
    #  'st_mtime': datetime.datetime(2023, 7, 7, 18, 55, 10, 755297),
    #  'st_birthtime': datetime.datetime(2023, 7, 7, 18, 55, 10, 754835),
    #  'st_name': 'com.apple.ibooks-sync.plist'}
    return FileInfo(
        name=info["st_name"],
        size=info["st_size"],
        ifmt=info["st_ifmt"],
        mtime=info["st_mtime"],
    )


def stat_file(afc: AfcService, path: str) -> FileInfo:
    info = afc.stat(path)
    info['st_name'] = posixpath.basename(path)
    return stat2fileinfo(info)


def byte2humansize(num_bytes):
    """
    Convert a size in bytes to a more human-readable format.

    :param num_bytes: Size in bytes.
    :return: Human-readable size.
    """
    if num_bytes < 1024.0:
        return f"{num_bytes}"
    for unit in ['K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        num_bytes /= 1024.0
        if num_bytes < 1024.0:
            return f"{num_bytes:3.1f}{unit}"
    return f"{num_bytes:.1f}Y"




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Mon Jan 15 2024 18:54:35 by codeskyblue
"""

from __future__ import annotations

import os
import sys

import click

from tidevice3.cli.cli_common import cli


@cli.command(name="exec", context_settings={"ignore_unknown_options": True})
@click.argument("args", nargs=-1)
@click.pass_context
def _exec(ctx: click.Context, args: list[str]):
    """ translate to pymobiledevice3 command, eg: t3 exec version """
    args = [sys.executable, '-m', 'pymobiledevice3'] + list(args)
    if ctx.obj['udid']:
        args += ['--udid', ctx.obj['udid']]
    os.execv(args[0], args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Mon Jan 08 2024 14:50:21 by codeskyblue
"""

import logging
import typing

import click
from pymobiledevice3.lockdown import LockdownClient

from tidevice3.api import screenshot, screenshot_png
from tidevice3.cli.cli_common import cli, pass_rsd

logger = logging.getLogger(__name__)


@cli.command("screenshot")
@click.argument("out", type=click.File("wb"))
@pass_rsd
def cli_screenshot(service_provider: LockdownClient, out: typing.BinaryIO):
    """get device screenshot"""
    if out.name.endswith(".png"):
        out.write(screenshot_png(service_provider))
    else:
        im = screenshot(service_provider)
        im.save(out)



# For iOS 17 auto start-tunnel

from __future__ import annotations

import logging
import os
import shlex
import shutil
import signal
import subprocess
import sys
import threading
import time
from typing import Mapping, NamedTuple, Tuple

import click
import fastapi
import uvicorn
from fastapi import FastAPI
from packaging.version import Version
from pymobiledevice3.exceptions import MuxException

from tidevice3.cli.cli_common import cli
from tidevice3.cli.list import list_devices
from tidevice3.utils.common import threadsafe_function

logger = logging.getLogger(__name__)


class Address(NamedTuple):
    ip: str
    port: int


def get_connected_devices() -> list[str]:
    """return list of udid"""
    try:
        devices = list_devices(usb=True, network=False)
    except MuxException as e:
        logger.error("list_devices failed: %s", e)
        return []
    return [d.Identifier for d in devices if Version(d.ProductVersion) >= Version("17")]


def guess_pymobiledevice3_path() -> str:
    pmd3path = shutil.which("pymobiledevice3")
    if not pmd3path:
        pmd3path = sys.executable + " -m pymobiledevice3"
    return pmd3path


class TunnelError(Exception):
    pass


@threadsafe_function
def start_tunnel(pmd3_path: str, udid: str) -> Tuple[Address, subprocess.Popen]:
    """
    Start program, should be killed when the main program quit

    Raises:
        TunnelError
    """
    # cmd = ["bash", "-c", "echo ::1 1234; sleep 10001"]
    log_prefix = f"[{udid}]"
    cmd = f"{pmd3_path} remote start-tunnel --script-mode --udid {udid}"
    logger.info("%s cmd: %s", log_prefix, cmd)
    process = subprocess.Popen(
        shlex.split(cmd), stdin=subprocess.DEVNULL, stdout=subprocess.PIPE
    )
    output_str = process.stdout.readline().decode("utf-8").strip()
    if output_str == "":
        raise TunnelError("pmd3 start-tunnel empty response")
    address, port_str = output_str.split()
    port = int(port_str)
    logger.info("%s tunnel address: %s", log_prefix, [address, port])
    process.stdout = subprocess.DEVNULL  # maybe not working
    return Address(address, port), process


class DeviceManager:
    def __init__(self):
        self.active_monitors: Mapping[str, subprocess.Popen] = {}
        self.running = True
        self.addresses: Mapping[str, Address] = {}
        self.pmd3_path = "pymobiledevice3"

    def update_devices(self):
        current_devices = set(get_connected_devices())
        active_udids = set(self.active_monitors.keys())

        # Start monitors for new devices
        for udid in current_devices - active_udids:
            self.active_monitors[udid] = None
            try:
                threading.Thread(name=f"{udid} keeper",
                                 target=self._start_tunnel_keeper,
                                 args=(udid,),
                                 daemon=True).start()
            except Exception as e:
                logger.error("udid: %s start-tunnel failed: %s", udid, e)

        # Stop monitors for disconnected devices
        for udid in active_udids - current_devices:
            logger.info("udid: %s quit, terminate related process", udid)
            process = self.active_monitors[udid]
            if process:
                process.terminate()
            self.active_monitors.pop(udid, None)
            self.addresses.pop(udid, None)

    def _start_tunnel_keeper(self, udid: str):
        while udid in self.active_monitors:
            try:
                addr, process = start_tunnel(self.pmd3_path, udid)
                self.active_monitors[udid] = process
                self.addresses[udid] = addr
                self._wait_process_exit(process, udid)
            except TunnelError:
                logger.exception("udid: %s start-tunnel failed", udid)
            time.sleep(3)
    
    def _wait_process_exit(self, process: subprocess.Popen, udid: str):
         while True:
            try:
                process.wait(1.0)
                self.addresses.pop(udid, None)
                logger.warning("udid: %s process exit with code: %s", udid, process.returncode)
                break
            except subprocess.TimeoutExpired:
                continue

    def shutdown(self):
        logger.info("terminate all processes")
        for process in self.active_monitors.values():
            if process:
                process.terminate()
        self.running = False

    def run_forever(self):
        while self.running:
            try:
                self.update_devices()
            except Exception as e:
                logger.exception("update_devices failed: %s", e)
            time.sleep(1)


@cli.command(context_settings={"show_default": True})
@click.option(
    "--pmd3-path",
    "pmd3_path",
    help="pymobiledevice3 cli path",
    default=guess_pymobiledevice3_path(),
)
@click.option("--port", "port", help="listen port", default=5555)
def tunneld(pmd3_path: str, port: int):
    """start server for iOS >= 17 auto start-tunnel, function like pymobiledevice3 remote tunneld"""
    if os.getuid() != 0:
        logger.error("Please run as root")
        sys.exit(1)

    manager = DeviceManager()
    app = FastAPI()

    @app.get("/")
    def get_devices():
        return manager.addresses

    @app.get("/shutdown")
    def shutdown():
        manager.shutdown()
        os.kill(os.getpid(), signal.SIGINT)
        return fastapi.Response(status_code=200, content="Server shutting down...")

    manager.pmd3_path = pmd3_path

    threading.Thread(
        target=manager.run_forever, daemon=True, name="device_manager"
    ).start()
    try:
        uvicorn.run(app, host="0.0.0.0", port=port)
    finally:
        logger.info("Shutting down...")
        manager.shutdown()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import datetime
import io
import logging
import time
from typing import Any, Iterator

import click
import imageio.v2 as imageio
import numpy as np
from PIL import Image, ImageDraw, ImageFont
from pymobiledevice3.lockdown import LockdownClient

from tidevice3.api import iter_screenshot
from tidevice3.cli.cli_common import cli, pass_rsd

logger = logging.getLogger(__name__)


def limit_fps(screenshot_iterator: Iterator[Any], fps: int, debug: bool = False) -> Iterator[Any]:
    """ Limit the frame rate of the screenshot iterator to the given FPS """
    frame_duration = 1.0 / fps
    next_frame_time = time.time()
    last_screenshot = None

    for screenshot in screenshot_iterator:
        current_time = time.time()

        if current_time >= next_frame_time:
            last_screenshot = screenshot

            # Write frame to video
            if debug:
                print(".", end="", flush=True)
            yield screenshot

            # Schedule next frame
            next_frame_time += frame_duration

        # Fill in with the last image if the next frame time is still in the future
        while next_frame_time <= current_time:
            if last_screenshot is not None:
                if debug:
                    print("o", end="", flush=True)
                yield last_screenshot
            next_frame_time += frame_duration


def draw_text(pil_img: Image.Image, text: str):
    """ GPT生成的，效果勉强吧，不太好，总比没有的强 """
    draw = ImageDraw.Draw(pil_img)
    font = ImageFont.load_default()

    # Calculate the bounding box of the text
    text_bbox = font.getbbox(text)
    text_width = text_bbox[2] - text_bbox[0]
    text_height = text_bbox[3] - text_bbox[1]
    text_x = 20
    text_y = 50

    # Define text color and background color
    text_color = (255, 0, 0)  # Red color
    background_color = (128, 128, 128, 128)  # Gray color with 50% transparency

    # Create a rectangle background for text
    background_rectangle = [
        (text_x - 10, text_y - 10),  # Upper left corner
        (text_x + text_width + 10, text_y + text_height + 10)  # Lower right corner
    ]
    draw.rectangle(background_rectangle, fill=background_color)
    draw.text((text_x, text_y), text, fill=text_color, font=font)
    return pil_img


def resize_for_ffmpeg(img: Image.Image) -> Image.Image:
    """
    部分机型截图的尺寸不对，所以需要resize，目前发现机型：iPhone x
    """
    w, h = img.size
    if w % 2 != 0:
        w -= 1
    if h % 2 != 0:
        h -= 1
    img = img.crop((0, 0, w, h))
    return img


@cli.command("screenrecord")
@click.option("--fps", default=5, help="frame per second")
@click.option("--show-time/--no-show-time", default=True, help="show time on screen")
@click.argument("out")
@pass_rsd
def cli_screenrecord(service_provider: LockdownClient, out: str, fps: int, show_time: bool):
    """ screenrecord to mp4 """
    writer = imageio.get_writer(out, fps=fps)
    frame_index = 0
    try:
        for png_data in limit_fps(iter_screenshot(service_provider), fps, debug=True):
            pil_img = Image.open(io.BytesIO(png_data))
            pil_img = resize_for_ffmpeg(pil_img)
            if show_time:
                time_str = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                draw_text(pil_img, f'Time: {time_str} Frame: {frame_index}')
            
            writer.append_data(np.array(pil_img))
            frame_index += 1
    except KeyboardInterrupt:
        print("")
    finally:
        writer.close()
        logger.info("screenrecord saved to %s", out)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Jan 05 2024 17:47:29 by codeskyblue
"""
import logging
import sys

import click
from pymobiledevice3.exceptions import NoDeviceConnectedError

from tidevice3.cli.cli_common import cli
from tidevice3.exceptions import BaseException, FatalError

logger = logging.getLogger(__name__)


def main():
    try:
        cli(auto_envvar_prefix='T3')
    except (FatalError, ValueError) as e:
        click.echo(f"Error: {e}")
        sys.exit(1)
    except NoDeviceConnectedError:
        logger.error("No device connected")
        sys.exit(1)
    except BaseException as e:
        click.echo(f"Error: {type(e)} {e}")
        sys.exit(1)
    except Exception as e:
        logger.exception("unhandled exception: %s", e)
        sys.exit(2)


if __name__ == '__main__':
    main()